"""State machine core."""

from enum import Enum
from functools import partial

from bluesky._vendor import six

from . import utils


NotSet = object()


class DefaultMeta(object):

    """Default configuration values."""

    states_enum_name = 'States'


class AttributeDict(dict):

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]


class StateMachineMetaclass(type):

    """Metaclass for state machine, to build all its logic."""

    def __new__(cls, name, bases, attrs):
        """Create state machine and add all logic and methods to it."""
        cls._set_up_context()
        new_class = super(cls, cls).__new__(cls, name, bases, attrs)
        cls.context.new_class = new_class

        parents = [b for b in bases if isinstance(b, cls)]
        if not parents:
            return cls.context.new_class

        cls._set_up_config_getter()
        cls._check_states_enum()
        cls._check_if_states_are_strings()
        cls._set_up_translator()
        cls._calculate_state_name()
        cls._check_state_value()
        cls._add_standard_attributes()
        cls._generate_standard_transitions()
        cls._generate_standard_methods()
        cls._generate_named_checkers()
        cls._generate_named_transitions()
        cls._add_new_methods()
        cls._set_complete_option()
        cls._complete_meta_for_new_class()

        new_class = cls.context.new_class
        del cls.context
        return new_class

    @classmethod
    def _set_up_context(cls):
        """Create context to keep all needed variables in."""
        cls.context = AttributeDict()
        cls.context.new_meta = {}
        cls.context.new_transitions = {}
        cls.context.new_methods = {}

    @classmethod
    def _check_states_enum(cls):
        """Check if states enum exists and is proper one."""
        states_enum_name = cls.context.get_config('states_enum_name')
        try:
            cls.context['states_enum'] = getattr(
                cls.context.new_class, states_enum_name)
        except AttributeError:
            raise ValueError('No states enum given!')

        proper = True
        try:
            if not issubclass(cls.context.states_enum, Enum):
                proper = False
        except TypeError:
            proper = False

        if not proper:
            raise ValueError(
                'Please provide enum instance to define available states.')

    @classmethod
    def _check_if_states_are_strings(cls):
        """Check if all states are strings."""
        for item in list(cls.context.states_enum):
            if not isinstance(item.value, six.string_types):
                raise ValueError(
                    'Item {name} is not string. Only strings are allowed.'
                    .format(name=item.name)
                )

    @classmethod
    def _check_state_value(cls):
        """Check initial state value - if is proper and translate it.

        Initial state is required.
        """
        state_value = cls.context.get_config('initial_state', None)
        state_value = state_value or getattr(
            cls.context.new_class, cls.context.state_name, None
        )

        if not state_value:
            raise ValueError(
                "Empty state is disallowed, yet no initial state is given!"
            )
        state_value = (
            cls.context
            .new_meta['translator']
            .translate(state_value)
        )
        cls.context.state_value = state_value

    @classmethod
    def _add_standard_attributes(cls):
        """Add attributes common to all state machines.

        These are methods for setting and checking state etc.

        """
        setattr(
            cls.context.new_class,
            cls.context.new_meta['state_attribute_name'],
            cls.context.state_value)
        setattr(
            cls.context.new_class,
            cls.context.state_name,
            utils.state_property)

        setattr(cls.context.new_class, 'is_', utils.is_)
        setattr(cls.context.new_class, 'can_be_', utils.can_be_)
        setattr(cls.context.new_class, 'set_', utils.set_)

    @classmethod
    def _generate_standard_transitions(cls):
        """Generate methods used for transitions."""
        allowed_transitions = cls.context.get_config('transitions', {})
        for key, transitions in allowed_transitions.items():
            key = cls.context.new_meta['translator'].translate(key)

            new_transitions = set()
            for trans in transitions:
                if not isinstance(trans, Enum):
                    trans = cls.context.new_meta['translator'].translate(trans)
                new_transitions.add(trans)

            cls.context.new_transitions[key] = new_transitions

        for state in cls.context.states_enum:
            if state not in cls.context.new_transitions:
                cls.context.new_transitions[state] = set()

    @classmethod
    def _generate_standard_methods(cls):
        """Generate standard setters, getters and checkers."""
        for state in cls.context.states_enum:
            getter_name = 'is_{name}'.format(name=state.value)
            cls.context.new_methods[getter_name] = utils.generate_getter(state)

            setter_name = 'set_{name}'.format(name=state.value)
            cls.context.new_methods[setter_name] = utils.generate_setter(state)

            checker_name = 'can_be_{name}'.format(name=state.value)
            checker = utils.generate_checker(state)
            cls.context.new_methods[checker_name] = checker

        cls.context.new_methods['actual_state'] = utils.actual_state
        cls.context.new_methods['as_enum'] = utils.as_enum
        cls.context.new_methods['force_set'] = utils.force_set

    @classmethod
    def _generate_named_checkers(cls):
        named_checkers = cls.context.get_config('named_checkers', None) or []
        for method, key in named_checkers:
            if method in cls.context.new_methods:
                raise ValueError(
                    "Name collision for named checker '{checker}' - this "
                    "name is reserved for other auto generated method."
                    .format(checker=method)
                )

            key = cls.context.new_meta['translator'].translate(key)
            cls.context.new_methods[method] = utils.generate_checker(key.value)

    @classmethod
    def _generate_named_transitions(cls):
        named_transitions = (
            cls.context.get_config('named_transitions', None) or [])

        translator = cls.context.new_meta['translator']
        for item in named_transitions:
            method, key, from_values = cls._unpack_named_transition_tuple(item)

            if method in cls.context.new_methods:
                raise ValueError(
                    "Name collision for transition '{transition}' - this name "
                    "is reserved for other auto generated method."
                    .format(transition=method)
                )

            key = translator.translate(key)
            cls.context.new_methods[method] = utils.generate_setter(key)

            if from_values:
                from_values = [translator.translate(k) for k in from_values]
                for s in cls.context.states_enum:
                    if s in from_values:
                        cls.context.new_transitions[s].add(key)

    @classmethod
    def _unpack_named_transition_tuple(cls, item):
        try:
            method, key = item
            from_values = cls.context['states_enum']
        except ValueError:
            method, key, from_values = item
            if from_values is None:
                from_values = []
            if not isinstance(from_values, list):
                from_values = list((from_values,))

        return method, key, from_values

    @classmethod
    def _add_new_methods(cls):
        """Add all generated methods to result class."""
        for name, method in cls.context.new_methods.items():
            if hasattr(cls.context.new_class, name):
                raise ValueError(
                    "Name collision in state machine class - '{name}'."
                    .format(name)
                )

            setattr(cls.context.new_class, name, method)

    @classmethod
    def _set_complete_option(cls):
        """Check and set complete option."""
        get_config = cls.context.get_config
        complete = get_config('complete', None)
        if complete is None:
            conditions = [
                get_config('transitions', False),
                get_config('named_transitions', False),
            ]
            complete = not any(conditions)

        cls.context.new_meta['complete'] = complete

    @classmethod
    def _set_up_config_getter(cls):
        meta = getattr(cls.context.new_class, 'Meta', DefaultMeta)
        cls.context.get_config = partial(get_config, meta)

    @classmethod
    def _set_up_translator(cls):
        translator = utils.EnumValueTranslator(cls.context['states_enum'])
        cls.context.new_meta['translator'] = translator

    @classmethod
    def _calculate_state_name(cls):
        cls.context.state_name = 'state'
        new_state_name = '_' + cls.context.state_name
        cls.context.new_meta['state_attribute_name'] = new_state_name

    @classmethod
    def _complete_meta_for_new_class(cls):
        cls.context.new_meta['transitions'] = cls.context.new_transitions
        cls.context.new_meta['config_getter'] = cls.context['get_config']
        setattr(cls.context.new_class, '_meta', cls.context['new_meta'])


class StateMachine(six.with_metaclass(StateMachineMetaclass)):

    """State machine."""


def get_config(original_meta, attribute, default=NotSet):
    for meta in [original_meta, DefaultMeta]:
        try:
            return getattr(meta, attribute)
        except AttributeError:
            pass

    if default is NotSet:
        raise

    return default



import os
import signal
import threading
import time
import time as ttime
from functools import partial

import pytest

import bluesky.plan_stubs as bps
from bluesky import FailedStatus, IllegalMessageSequence, Msg, RunEngineInterrupted
from bluesky.callbacks.mpl_plotting import LivePlot
from bluesky.examples import (
    checkpoint_forever,
    conditional_break,
    conditional_pause,
    fly_gen,
    simple_scan,
    simple_scan_saving,
    sleepy,
    stepscan,
    wait_multiple,
    wait_one,
)
from bluesky.tests import uses_os_kill_sigint

from .utils import _careful_event_set, _fabricate_asycio_event


def test_msgs(hw):
    m = Msg("set", hw.motor, {"motor": 5})
    assert m.command == "set"
    assert m.obj is hw.motor
    assert m.args == ({"motor": 5},)
    assert m.kwargs == {}

    m = Msg("read", hw.motor)
    assert m.command == "read"
    assert m.obj is hw.motor
    assert m.args == tuple()  # noqa: C408
    assert m.kwargs == {}

    m = Msg("create", name="primary")
    assert m.command == "create"
    assert m.obj is None
    assert m.args == tuple()  # noqa: C408
    assert m.kwargs == {"name": "primary"}

    m = Msg("sleep", None, 5)
    assert m.command == "sleep"
    assert m.obj is None
    assert m.args == (5,)
    assert m.kwargs == {}


def run(RE, gen, *args, **kwargs):
    assert RE.state == "idle"
    RE(gen(*args, **kwargs))
    assert RE.state == "idle"


def test_simple(RE, hw):
    run(RE, simple_scan, hw.motor)


def test_conditional_break(RE, hw):
    run(RE, conditional_break, hw.det, hw.motor, 0.2)


def test_sleepy(RE, hw):
    run(RE, sleepy, hw.det, hw.motor)


def test_wait_one(RE, hw):
    run(RE, wait_one, hw.det, hw.motor)


def test_wait_multiple(RE, hw):
    run(RE, wait_multiple, hw.det, [hw.motor1, hw.motor2])


def test_hard_pause(RE, hw):
    assert RE.state == "idle"
    with pytest.raises(RunEngineInterrupted):
        RE(conditional_pause(hw.det, hw.motor, False, True))
    assert RE.state == "paused"
    with pytest.raises(RunEngineInterrupted):
        RE.resume()
    assert RE.state == "paused"
    RE.abort()
    assert RE.state == "idle"


def test_deferred_pause(RE):
    # deferred pause should be processed once and then clear
    # (future checkpoints should not trigger another pause)
    assert RE.state == "idle"
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("pause", defer=True), Msg("checkpoint"), Msg("checkpoint"), Msg("checkpoint")])
    assert RE.state == "paused"
    RE.resume()
    assert RE.state == "idle"


def test_deferred_pause1(RE):
    # deferred pause should never be processed, being superceded by a hard
    # pause
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("pause", defer=True), Msg("pause", defer=False), Msg("checkpoint")])
    assert RE.state == "paused"
    RE.resume()
    assert RE.state == "idle"


def test_deferred_pause2(RE):
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("pause", defer=True), Msg("checkpoint"), Msg("pause", defer=True), Msg("checkpoint")])
    assert RE.state == "paused"
    with pytest.raises(RunEngineInterrupted):
        RE.resume()
    assert RE.state == "paused"
    RE.resume()
    assert RE.state == "idle"


def test_hard_pause_no_checkpoint(RE):
    assert RE.state == "idle"
    with pytest.raises(RunEngineInterrupted):
        (RE([Msg("clear_checkpoint"), Msg("pause", False)]),)
    assert RE.state == "idle"


def test_deferred_pause_no_checkpoint(RE):
    assert RE.state == "idle"
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("clear_checkpoint"), Msg("pause", True)])
    assert RE.state == "idle"


def test_pause_from_outside(RE):
    assert RE.state == "idle"

    def local_pause(delay):
        time.sleep(delay)
        RE.request_pause()

    th = threading.Thread(target=partial(local_pause, 1))
    th.start()
    with pytest.raises(RunEngineInterrupted):
        RE(checkpoint_forever())
    assert RE.state == "paused"

    # Cue up a second pause requests in 2 seconds.
    th = threading.Thread(target=partial(local_pause, 2))
    th.start()
    with pytest.raises(RunEngineInterrupted):
        RE.resume()
    assert RE.state == "paused"

    RE.abort()
    assert RE.state == "idle"


def test_simple_scan_saving(RE, hw):
    run(RE, simple_scan_saving, hw.det, hw.motor)


def print_event_time(name, doc):
    print("===== EVENT TIME:", doc["time"], "=====")


def test_calltime_subscription(RE, hw):
    assert RE.state == "idle"
    RE(simple_scan_saving(hw.det, hw.motor), {"event": print_event_time})
    assert RE.state == "idle"


def test_stateful_subscription(RE, hw):
    assert RE.state == "idle"
    token = RE.subscribe(print_event_time, "event")
    RE(simple_scan_saving(hw.det, hw.motor))
    RE.unsubscribe(token)
    assert RE.state == "idle"


def test_live_plotter(RE, hw):
    RE.ignore_callback_exceptions = False
    try:
        import matplotlib.pyplot as plt

        del plt
    except ImportError as ie:
        pytest.skip(f"Skipping live plot test because matplotlib is not installed.Error was: {ie}")

    my_plotter = LivePlot("det", "motor")
    assert RE.state == "idle"
    RE(stepscan(hw.det, hw.motor), {"all": my_plotter})
    assert RE.state == "idle"
    xlen = len(my_plotter.x_data)
    assert xlen > 0
    ylen = len(my_plotter.y_data)
    assert xlen == ylen
    RE.ignore_callback_exceptions = True


def test_sample_md_dict_requirement(RE, hw):
    # We avoid a json ValidationError and make a user-friendly ValueError.
    with pytest.raises(ValueError):
        RE(simple_scan(hw.motor), sample=1)
    RE(simple_scan(hw.motor), sample={"number": 1})  # should not raise
    RE(simple_scan(hw.motor), sample="label")  # should not raise


def test_md_dict(RE, hw):
    _md({}, RE, hw)


def test_md_historydict(RE, hw):
    try:
        import historydict
    except ImportError as ie:
        pytest.skip(f"Skipping test because historydict cannot be imported. Error was {ie}")
    _md(historydict.HistoryDict(":memory:"), RE, hw)


def _md(md, RE, hw):
    RE.ignore_callback_exceptions = False

    # Check persistence.
    scan = simple_scan(hw.motor)
    RE(scan, project="sitting")
    # 'project' should not persist
    scan = simple_scan(hw.motor)
    RE(scan, {"start": [validate_dict_cb_opposite("project")]})
    # ...unless we add it to RE.md
    RE.md["project"] = "sitting"
    scan = simple_scan(hw.motor)
    RE(scan, {"start": [validate_dict_cb("project", "sitting")]})
    # new values to 'project' passed in the call override the value in md
    scan = simple_scan(hw.motor)
    RE(scan, {"start": [validate_dict_cb("project", "standing")]}, project="standing")
    # ...but they do not update the value in md
    assert RE.md["project"] == "sitting"


def validate_dict_cb(key, val):
    def callback(name, doc):
        assert key in doc
        assert doc[key] == val

    return callback


def validate_dict_cb_opposite(key):
    def callback(name, doc):
        assert key not in doc

    return callback


def test_simple_fly(RE, hw):
    hw.flyer1.loop = RE.loop
    RE(fly_gen(hw.flyer1))


def test_list_of_msgs(RE, hw):
    # smoke tests checking that RunEngine accepts a plain list of Messages
    RE([Msg("open_run"), Msg("set", hw.motor, 5), Msg("close_run")])


def test_suspend(RE, hw):
    ev = _fabricate_asycio_event(RE.loop)

    test_list = [
        Msg("open_run"),
        Msg("checkpoint"),
        Msg("sleep", None, 0.2),
        Msg("set", hw.motor, 5),
        Msg("trigger", hw.det),
        Msg("declare_stream", None, hw.motor, hw.det, name="primary"),
        Msg("create", name="primary"),
        Msg("read", hw.motor),
        Msg("read", hw.det),
        Msg("save"),
        Msg("close_run"),
    ]
    assert RE.state == "idle"

    def resume_cb():
        RE.loop.call_soon_threadsafe(_careful_event_set(ev))

    def local_suspend():
        RE.request_suspend(ev.wait)
        # wait a second and then resume
        threading.Timer(1, resume_cb).start()

    out = []

    def ev_cb(name, ev):
        out.append(ev)

    # trigger the suspend right after the check point
    threading.Timer(0.1, local_suspend).start()
    # grab the start time
    start = ttime.time()
    # run, this will not return until it is done
    RE(test_list, {"event": ev_cb})
    # check to make sure it took long enough
    assert out[0]["time"] - start > 1.1

    assert RE.state == "idle"


@uses_os_kill_sigint
def test_pause_resume(RE):
    from bluesky.utils import ts_msg_hook

    RE.msg_hook = ts_msg_hook
    ev = _fabricate_asycio_event(RE.loop)

    def done():
        print("Done")
        RE.loop.call_soon_threadsafe(_careful_event_set(ev))

    pid = os.getpid()

    def sim_kill():
        os.kill(pid, signal.SIGINT)

    scan = [
        Msg("checkpoint"),
        Msg(
            "wait_for",
            None,
            [
                ev.wait,
            ],
        ),
    ]
    assert RE.state == "idle"
    start = ttime.time()
    threading.Timer(1, sim_kill).start()
    threading.Timer(1.5, sim_kill).start()
    threading.Timer(2, done).start()

    with pytest.raises(RunEngineInterrupted):
        RE(scan)
    assert RE.state == "paused"
    mid = ttime.time()
    RE.resume()
    assert RE.state == "idle"
    stop = ttime.time()

    time.sleep(3)
    assert mid - start > 1
    assert stop - start > 2


@uses_os_kill_sigint
def test_pause_abort(RE):
    ev = _fabricate_asycio_event(RE.loop)

    def done():
        print("Done")
        try:
            RE.loop.call_soon_threadsafe(_careful_event_set(ev))
        except RuntimeError:
            ...

    pid = os.getpid()

    def sim_kill():
        os.kill(pid, signal.SIGINT)

    scan = [
        Msg("checkpoint"),
        Msg(
            "wait_for",
            None,
            [
                ev.wait,
            ],
        ),
    ]
    assert RE.state == "idle"
    start = ttime.time()
    threading.Timer(0.1, sim_kill).start()
    threading.Timer(0.2, sim_kill).start()
    threading.Timer(1, done).start()

    with pytest.raises(RunEngineInterrupted):
        RE(scan)
    assert RE.state == "paused"
    mid = ttime.time()
    RE.abort()
    assert RE.state == "idle"
    stop = ttime.time()

    assert mid - start > 0.1
    assert stop - start < 1


@uses_os_kill_sigint
def test_abort(RE):
    ev = _fabricate_asycio_event(RE.loop)

    def done():
        print("Done")
        try:
            RE.loop.call_soon_threadsafe(_careful_event_set(ev))
        except RuntimeError:
            ...

    pid = os.getpid()

    def sim_kill():
        os.kill(pid, signal.SIGINT)

    scan = [
        Msg("checkpoint"),
        Msg(
            "wait_for",
            None,
            [
                ev.wait,
            ],
        ),
    ]
    assert RE.state == "idle"
    start = ttime.time()
    threading.Timer(0.1, sim_kill).start()
    threading.Timer(0.2, sim_kill).start()
    threading.Timer(0.4, done).start()
    with pytest.raises(RunEngineInterrupted):
        RE(scan)
    stop = ttime.time()

    assert RE.state == "paused"
    assert stop - start < 0.4
    RE.abort()
    assert RE.state == "idle"


def test_rogue_sigint(RE):
    def bad_scan():
        yield Msg("open_run")
        yield Msg("checkpoint")
        raise KeyboardInterrupt()

    with pytest.raises(RunEngineInterrupted):
        RE(bad_scan())
    assert RE.state == "idle"


def test_seqnum_nonrepeated(RE, hw):
    def gen():
        yield Msg("open_run")
        yield Msg("declare_stream", None, hw.motor, name="primary")
        yield Msg("create", name="primary")
        yield Msg("set", hw.motor, 1)
        yield Msg("read", hw.motor)
        yield Msg("save")
        yield Msg("checkpoint")
        yield Msg("create", name="primary")
        yield Msg("set", hw.motor, 2)
        yield Msg("read", hw.motor)
        yield Msg("save")
        yield Msg("pause")
        yield Msg("create", name="primary")
        yield Msg("set", hw.motor, 3)
        yield Msg("read", hw.motor)
        yield Msg("save")
        yield Msg("close_run")

    seq_nums = []

    def f(name, doc):
        seq_nums.append(doc["seq_num"])

    RE.verbose = True

    with pytest.raises(RunEngineInterrupted):
        RE(gen(), {"event": f})

    print("RESUMING!!!!")
    RE.resume()
    assert seq_nums == [1, 2, 2, 3]


def test_duplicate_keys(RE, hw):
    # two detectors, same data keys

    def gen():
        yield (Msg("open_run"))
        yield Msg("declare_stream", None, hw.det, hw.identical_det, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("trigger", hw.det))
        yield (Msg("trigger", hw.identical_det))
        yield (Msg("read", hw.det))
        yield (Msg("read", hw.identical_det))
        yield (Msg("save"))

    with pytest.raises(ValueError):
        RE(gen())


def test_illegal_sequences(RE, hw):
    def gen1():
        # two 'create' msgs in a row
        yield (Msg("open_run"))
        yield Msg("declare_stream", None, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("create", name="primary"))
        yield (Msg("close_run"))

    with pytest.raises(IllegalMessageSequence):
        RE(gen1())

    def gen2():
        # two 'save' msgs in a row
        yield (Msg("open_run"))
        yield Msg("declare_stream", None, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("save"))
        yield (Msg("save"))
        yield (Msg("close_run"))

    with pytest.raises(IllegalMessageSequence):
        RE(gen2())

    def gen3():
        # 'configure' after 'create', before 'save'
        yield (Msg("open_run"))
        yield Msg("declare_stream", None, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("configure", hw.motor, {}))

    with pytest.raises(IllegalMessageSequence):
        RE(gen3())

    def gen4():
        # two 'drop' msgs in a row
        yield (Msg("open_run"))
        yield Msg("declare_stream", None, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("drop"))
        yield (Msg("drop"))
        yield (Msg("close_run"))

    with pytest.raises(IllegalMessageSequence):
        RE(gen4())


def test_new_ev_desc(RE, hw):
    descs = []

    def collect_descs(name, doc):
        descs.append(doc)

    def gen1():
        # configure between two events -> two descs
        yield (Msg("open_run"))
        yield Msg("declare_stream", None, hw.motor, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("read", hw.motor))
        yield (Msg("save"))
        yield (Msg("configure", hw.motor, {}))
        yield (Msg("create", name="primary"))
        yield (Msg("read", hw.motor))
        yield (Msg("save"))
        yield (Msg("close_run"))

    descs.clear()
    RE(gen1(), {"descriptor": collect_descs})
    assert len(descs) == 2

    def gen2():
        # configure between two events and explicitly before any events
        # -> two descs
        yield (Msg("open_run"))
        yield (Msg("configure", hw.motor, {}))
        yield Msg("declare_stream", None, hw.motor, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("read", hw.motor))
        yield (Msg("save"))
        yield (Msg("configure", hw.motor, {}))
        yield (Msg("create", name="primary"))
        yield (Msg("read", hw.motor))
        yield (Msg("save"))
        yield (Msg("close_run"))

    descs.clear()
    RE(gen2(), {"descriptor": collect_descs})
    assert len(descs) == 2

    def gen3():
        # configure once before any events -> one desc
        yield (Msg("open_run"))
        yield (Msg("configure", hw.motor, {}))
        yield Msg("declare_stream", None, hw.motor, name="primary")
        yield (Msg("create", name="primary"))
        yield (Msg("read", hw.motor))
        yield (Msg("save"))
        yield (Msg("create", name="primary"))
        yield (Msg("read", hw.motor))
        yield (Msg("save"))
        yield (Msg("close_run"))

    descs.clear()
    RE(gen3(), {"descriptor": collect_descs})
    assert len(descs) == 1


def test_clear_checkpoint(RE):
    bad_plan = [Msg("checkpoint"), Msg("clear_checkpoint"), Msg("pause"), "lies"]
    good_plan = [Msg("pause")]
    fine_plan = [Msg("clear_checkpoint")]

    with pytest.raises(RunEngineInterrupted):
        RE(good_plan)
    assert RE.state == "paused"
    RE.stop()

    RE(fine_plan)
    assert RE.state == "idle"

    # this should raise an attribute error if the last entry in the plan
    # is passed to the run engine (but that should not happen because it should
    # die when it hits the 'pause' and has no checkpoint)
    with pytest.raises(RunEngineInterrupted):
        RE(bad_plan)
    assert RE.state == "idle"


def test_interruption_exception(RE):
    with pytest.raises(RunEngineInterrupted):
        RE([Msg("checkpoint"), Msg("pause")])
    RE.stop()


def test_failed_status_object(RE):
    try:
        from ophyd import StatusBase
    except ImportError:
        pytest.xfail("No ophyd")

    class failer:
        def set(self, inp):
            st = StatusBase()
            threading.Timer(1, st._finished, kwargs=dict(success=False)).start()  # noqa: C408
            return st

        def trigger(self):
            st = StatusBase()
            threading.Timer(1, st._finished, kwargs=dict(success=False)).start()  # noqa: C408
            return st

        def stop(self, *, success=False):
            pass

    ff = failer()
    with pytest.raises(FailedStatus):
        RE([Msg("set", ff, None, group="a"), Msg("wait", None, group="a")])

    with pytest.raises(FailedStatus):
        RE([Msg("trigger", ff, group="a"), Msg("wait", None, group="a")])


def test_rewindable_by_default(RE):
    class Sentinel(Exception):
        pass

    def plan():
        yield Msg("pause")
        raise Sentinel

    with pytest.raises(RunEngineInterrupted):
        RE(plan())
    assert RE.state == "paused"
    with pytest.raises(Sentinel):
        RE.resume()

    def plan():
        yield Msg("clear_checkpoint")
        yield Msg("pause")

    with pytest.raises(RunEngineInterrupted):
        RE(plan())  # cannot pause
    assert RE.state == "idle"


def test_sync_trigger_delay(hw):
    def _time_test(f, t, *args, **kwargs):
        start = ttime.time()
        st = f(*args, **kwargs)
        while not st.done:
            ttime.sleep(0.01)
        end = ttime.time()
        assert (end - start) > t

    hw.motor.delay = 0.5
    hw.det.exposure_time = 0.5
    assert hw.det.exposure_time == 0.5

    _time_test(hw.motor.set, 0.5, 1)
    _time_test(hw.det.trigger, 0.5)


def test_async_trigger_delay(RE, hw):
    hw.motor.loop = RE.loop
    hw.det.loop = RE.loop

    def _time_test(f, t, *args, **kwargs):
        start = ttime.time()
        RE(f(*args, **kwargs))
        end = ttime.time()
        assert (end - start) > t

    hw.motor.delay = 0.5
    hw.det.exposure_time = 0.5

    _time_test(bps.trigger, 0.5, hw.det, wait=True)
    _time_test(bps.abs_set, 0.5, hw.motor, 1, wait=True)



import collections
import functools
import operator
from enum import Enum

import numpy as np
from cycler import cycler

try:
    # cytools is a drop-in replacement for toolz, implemented in Cython
    from cytools import partition
except ImportError:
    from toolz import partition

from .utils import is_movable, snake_cyclers


def spiral(x_motor, y_motor, x_start, y_start, x_range, y_range, dr, nth, *, dr_y=None, tilt=0.0):
    """Spiral scan, centered around (x_start, y_start)

    Parameters
    ----------
    x_motor : object, optional
        any 'setable' object (motor, temp controller, etc.)
    y_motor : object, optional
        any 'setable' object (motor, temp controller, etc.)
    x_start : float
        x center
    y_start : float
        y center
    x_range : float
        x width of spiral
    y_range : float
        y width of spiral
    dr : float
        Delta radius along the minor axis of the ellipse.
    dr_y : float, optional
        Delta radius along the major axis of the ellipse, if not specifed
        defaults to dr
    nth : float
        Number of theta steps
    tilt : float, optional
        Tilt angle in radians, default 0.0

    Returns
    -------
    cyc : cycler
    """
    if dr_y is None:
        dr_aspect = 1
    else:
        dr_aspect = dr_y / dr

    half_x = x_range / 2
    half_y = y_range / (2 * dr_aspect)

    r_max = np.sqrt(half_x**2 + half_y**2)
    num_ring = 1 + int(r_max / dr)
    tilt_tan = np.tan(tilt + np.pi / 2.0)

    x_points, y_points = [], []

    for i_ring in range(1, num_ring + 2):
        radius = i_ring * dr
        angle_step = 2.0 * np.pi / (i_ring * nth)

        for i_angle in range(int(i_ring * nth)):
            angle = i_angle * angle_step
            x = radius * np.cos(angle)
            y = radius * np.sin(angle) * dr_aspect
            if (abs(x - (y / dr_aspect) / tilt_tan) <= half_x) and (abs(y / dr_aspect) <= half_y):
                x_points.append(x_start + x)
                y_points.append(y_start + y)

    cyc = cycler(x_motor, x_points)
    cyc += cycler(y_motor, y_points)
    return cyc


def spiral_square_pattern(x_motor, y_motor, x_center, y_center, x_range, y_range, x_num, y_num):
    """
    Square spiral scan, centered around (x_start, y_start)

    Parameters
    ----------
    x_motor : object, optional
        any 'setable' object (motor, temp controller, etc.)
    y_motor : object, optional
        any 'setable' object (motor, temp controller, etc.)
    x_center : float
        x center
    y_center : float
        y center
    x_range : float
        x width of spiral
    y_range : float
        y width of spiral
    x_num : float
        number of x axis points
    y_num : float
        number of y axis points
    Returns
    -------
    cyc : cycler
    """
    x_points, y_points = [], []

    # checks if x_num/y_num is even or odd and sets the required offset
    # parameter for the start point from the centre point.
    if x_num % 2 == 0:
        x_offset = 0.5
    else:
        x_offset = 0

    if y_num % 2 == 0:
        y_offset = -0.5
    else:
        y_offset = 0

    num_ring = max(x_num, y_num)
    x_delta = x_range / (x_num - 1)
    y_delta = y_range / (y_num - 1)

    # include the first point, as it is the first 'ring' to include.
    x_points.append(x_center - x_delta * x_offset)
    y_points.append(y_center - y_delta * y_offset)

    # set the number of found points to 0
    num_pnts_fnd = 1

    # step through each of the rings required to map out the entire area.
    for i_ring in range(2, num_ring + 1, 1):
        # step through each of the 'sides' of the ring if the constant value
        # for each side is within the range to plot and
        # that we have not already found all the required points.
        # SIDE 1
        if (abs(i_ring - 1 - x_offset) <= x_num / 2) and (num_pnts_fnd < x_num * y_num):
            for n in range(i_ring - 2, -i_ring, -1):
                # Ensure that the variable value for this side is within the
                # range to plot and that we have not already
                # found all the required points.
                if (abs(n - y_offset) < y_num / 2) and (num_pnts_fnd < y_num * x_num):
                    x = x_center - x_delta * x_offset + x_delta * (i_ring - 1)
                    y = y_center - y_delta * y_offset + y_delta * n
                    num_pnts_fnd += 1

                    x_points.append(x)
                    y_points.append(y)

        # SIDE 2
        if (abs(-i_ring + 1 - y_offset) < y_num / 2) and (num_pnts_fnd < x_num * y_num):
            for n in range(i_ring - 2, -i_ring, -1):
                # Ensure that the variable value for this side is within the
                # range to plot and that we have not already
                # found all the required points.
                if (abs(n - x_offset) < x_num / 2) and (num_pnts_fnd < y_num * x_num):
                    x = x_center - x_delta * x_offset + x_delta * n
                    y = y_center - y_delta * y_offset + y_delta * (-i_ring + 1)

                    num_pnts_fnd += 1

                    x_points.append(x)
                    y_points.append(y)

        # SIDE 3
        if (abs(-i_ring + 1 - x_offset) < x_num / 2) and (num_pnts_fnd < x_num * y_num):
            for n in range(-i_ring + 2, i_ring, 1):
                # Ensure that the variable value for this side is within the
                # range to plot and that we have not already
                # found all the required points.
                if (abs(n - y_offset) < y_num / 2) and (num_pnts_fnd < y_num * x_num):
                    x = x_center - x_delta * x_offset + x_delta * (-i_ring + 1)
                    y = y_center - y_delta * y_offset + y_delta * n
                    num_pnts_fnd += 1

                    x_points.append(x)
                    y_points.append(y)

        # SIDE 4
        if (abs(i_ring - 1 - y_offset) < y_num / 2) and (num_pnts_fnd < x_num * y_num):
            for n in range(-i_ring + 2, i_ring, 1):
                # Ensure that the variable value for this side is within the
                # range to plot and that we have not already
                # found all the required points.
                if (abs(n - x_offset) < x_num / 2) and (num_pnts_fnd < y_num * x_num):
                    x = x_center - x_delta * x_offset + x_delta * n
                    y = y_center - y_delta * y_offset + y_delta * (i_ring - 1)

                    num_pnts_fnd += 1

                    x_points.append(x)
                    y_points.append(y)

    cyc = cycler(x_motor, x_points)
    cyc += cycler(y_motor, y_points)

    return cyc


def spiral_fermat(x_motor, y_motor, x_start, y_start, x_range, y_range, dr, factor, *, dr_y=None, tilt=0.0):
    """Absolute fermat spiral scan, centered around (x_start, y_start)

    Parameters
    ----------
    x_motor : object, optional
        any 'setable' object (motor, temp controller, etc.)
    y_motor : object, optional
        any 'setable' object (motor, temp controller, etc.)
    x_start : float
        x center
    y_start : float
        y center
    x_range : float
        x width of spiral
    y_range : float
        y width of spiral
    dr : float
        delta radius along the minor axis of the ellipse.
    dr_y : float, optional
        Delta radius along the major axis of the ellipse, if not specifed defaults to dr
    factor : float
        radius gets divided by this
    tilt : float, optional
        Tilt angle in radians, default 0.0

    Returns
    -------
    cyc : cycler
    """
    if dr_y is None:
        dr_aspect = 1
    else:
        dr_aspect = dr_y / dr

    phi = 137.508 * np.pi / 180.0

    half_x = x_range / 2
    half_y = y_range / (2 * dr_aspect)
    tilt_tan = np.tan(tilt + np.pi / 2.0)

    x_points, y_points = [], []

    diag = np.sqrt(half_x**2 + half_y**2)
    num_rings = int((1.5 * diag / (dr / factor)) ** 2)
    for i_ring in range(1, num_rings):
        radius = np.sqrt(i_ring) * dr / factor
        angle = phi * i_ring
        x = radius * np.cos(angle)
        y = radius * np.sin(angle) * dr_aspect

        if (abs(x - (y / dr_aspect) / tilt_tan) <= half_x) and (abs(y) <= half_y):
            x_points.append(x_start + x)
            y_points.append(y_start + y)

    cyc = cycler(x_motor, x_points)
    cyc += cycler(y_motor, y_points)
    return cyc


def inner_list_product(args):
    """Scan over one multi-motor trajectory.

    Parameters
    ----------
    args : list
        patterned like (``motor1, position_list1,``
                        ``...,``
                        ``motorN, position_listN``)
        Motors can be any 'settable' object (motor, temp controller, etc.)
        ``position_list``'s are lists of positions, all lists must have the
        same length.
    Returns
    -------
    cyc : cycler
    """
    if len(args) % 2 != 0:
        raise ValueError("Wrong number of positional arguments for 'inner_list_product'")

    cyclers = []
    for (
        motor,
        pos_list,
    ) in partition(2, args):
        c = cycler(motor, pos_list)
        cyclers.append(c)
    return functools.reduce(operator.add, cyclers)


def outer_list_product(args, snake_axes):
    """Scan over a mesh; each motor is on an independent trajectory.

        Parameters
        ----------
        args
            patterned like (``motor1, position_list1,``
                            ``motor2, position_list2,``
                            ``motor3, position_list3,``
                            ``...,``
                            ``motorN, position_listN``)

            The first motor is the "slowest", the outer loop. ``position_list``'s
            are lists of positions, all lists must have the same length.
    .
        snake_axes
            which axes should be snaked, either ``False`` (do not snake any axes),
            ``True`` (snake all axes) or a list of axes to snake. "Snaking" an axis
            is defined as following snake-like, winding trajectory instead of a
            simple left-to-right trajectory.


        See Also
        --------
        :func:`bluesky.plan_patterns.inner_list_product`

        Returns
        -------
        cyc : cycler
    """
    snaking = []
    cyclers = []
    for motor, pos_list in partition(2, args):
        if not snake_axes:
            snaking.append(False)
        elif isinstance(snake_axes, collections.abc.Iterable):
            if motor in snake_axes:
                snaking.append(True)
            else:
                snaking.append(False)
        elif snake_axes:
            if not snaking:
                snaking.append(False)
            else:
                snaking.append(True)
        else:
            raise ValueError(
                "The snake_axes arg to ``outer_list_product`` "
                'must be either "False" (do not snake any axes), '
                '"True" (snake all axes) or a list of axes to '
                f"snake. Instead it is {snake_axes}."
            )

        c = cycler(motor, pos_list)
        cyclers.append(c)

    return snake_cyclers(cyclers, snaking)


def inner_product(num, args):
    """Scan over one multi-motor trajectory.

    Parameters
    ----------
    num : integer
        number of steps
    args : list of {Positioner, Positioner, int}
        patterned like (``motor1, start1, stop1, ..., motorN, startN, stopN``)
        Motors can be any 'setable' object (motor, temp controller, etc.)

    Returns
    -------
    cyc : cycler
    """
    if len(args) % 3 != 0:
        raise ValueError("Wrong number of positional arguments for 'inner_product'")

    cyclers = []
    for (
        motor,
        start,
        stop,
    ) in partition(3, args):
        steps = np.linspace(start, stop, num=num, endpoint=True)
        c = cycler(motor, steps)
        cyclers.append(c)
    return functools.reduce(operator.add, cyclers)


class OuterProductArgsPattern(Enum):
    PATTERN_1 = 1
    PATTERN_2 = 2


def classify_outer_product_args_pattern(args):
    """
    Classifies the pattern of grid scan arguments in the list `args`.
    Checks the argument list for consistency, in particular checks
    to location of movable objects (motors) in the list.
    Should be used together with the function `chunk_outer_product_args`.

    Parameters
    ----------
    args: iterable
        The list of grid scan arguments. Two pattern of arguments
        are supported. See the description of the identical parameter
        for the `chunk_outer_product_args`.

    Returns
    -------
    pattern: OuterProductArgsPattern
        Detected pattern

    Raises
    ------
    ValueError is raised if the pattern can not be identified or the list
    is inconsistent.
    """

    args = list(args)
    pattern = None

    def _verify_motor_locations(args, pattern):
        # Verify the motors are preset only at correct positions in the list
        if pattern == OuterProductArgsPattern.PATTERN_1:
            # Positions of the movable objects (motors)
            pos_movable = list(range(0, len(args), 4))
        elif pattern == OuterProductArgsPattern.PATTERN_2:
            # Positions of the movable objects (motors)
            pos_movable = [0] + list(range(4, len(args), 5))
        else:
            raise ValueError(f"Unknown pattern '{pattern}'")
        for n, element in enumerate(args):
            # Check if the element is the motor
            flag = is_movable(element)
            # If the element is expected to be the motor, then flip the flag
            if n in pos_movable:
                flag = not flag
            # Now the flag is True if the motor is out of place in the list of arguments
            if flag:
                return False
        return True

    # div_4 - the correct number of elements for pattern 1, div_5 - for pattern 2
    div_4, div_5 = not (len(args) % 4), (len(args) > 4) and not ((len(args) - 4) % 5)

    # Check the number of elements in 'args'
    if not div_4 and not div_5:
        raise ValueError(f"Wrong number of elements in 'args': len(args) = {len(args)}")

    args_valid = False
    if div_4 and not div_5:
        pattern = OuterProductArgsPattern.PATTERN_1
        args_valid = _verify_motor_locations(args, pattern)
    elif not div_4 and div_5:
        pattern = OuterProductArgsPattern.PATTERN_2
        args_valid = _verify_motor_locations(args, pattern)
    else:
        for p in OuterProductArgsPattern:
            if _verify_motor_locations(args, p):
                pattern = p
                args_valid = True
                break

    if not args_valid:
        raise ValueError(
            f"Incorrect order of elements in the argument list 'args': "
            f"some of the movable objects (motors) are out of place "
            f"(args = {args})"
        )

    return pattern


def chunk_outer_product_args(args, pattern=None):
    """Scan over a mesh; each motor is on an independent trajectory.

    Parameters
    ----------
    args: iterable
        Two patterns are supported:

        Pattern 1: (``motor1, start1, stop1, num1,```
                    ``motor2, start2, stop2, num2,``
                    ``motor3, start3, stop3, num3,`` ...
                    ``motorN, startN, stopN, numN``)

        Pattern 2: (``motor1, start1, stop1, num1,```
                    ``motor2, start2, stop2, num2, snake2,``
                    ``motor3, start3, stop3, num3, snake3,`` ...
                    ``motorN, startN, stopN, numN, snakeN``)

        All elements 'motorX' must be movable objects. There must be no
        movable objects in the other positions in the list.

        In Pattern 2, the first motor is the "slowest", the outer loop. For all motors
        except the first motor, there is a "snakeX" argument: a boolean
        indicating whether to following snake-like, winding trajectory or a
        simple left-to-right trajectory.
    pattern: OuterProductArgsPattern
        If pattern of 'args' is known, then it can be explicitely specified.
        In this case the automated recognition of the pattern is not performed
        and consistency of the argument list is not verified. Use
        `classify_outer_product_args_pattern` to verify consistency of `args`.

    See Also
    --------
    `bluesky.plan_patterns.outer_product`
    `bluesky.plan_patterns.classify_outer_product_args_pattern`

    Yields
    ------
    (motor, start, stop, num, snake)

    The `snake` value is always `False` for Pattern 1
    """

    if pattern is None:
        pattern = classify_outer_product_args_pattern(args)
    else:
        if not isinstance(pattern, OuterProductArgsPattern):
            raise ValueError(f"The parameter 'pattern' must have type OuterProductArgsPattern: {type(pattern)} ")

    args = list(args)

    if pattern == OuterProductArgsPattern.PATTERN_1:
        # Set 'snaked' to False for every motor
        for n in range(1, int(len(args) / 4) + 1):
            args.insert(5 * n - 1, False)
    elif pattern == OuterProductArgsPattern.PATTERN_2:
        # The first (slowest) axis is never "snaked." Insert False to
        # make it easy to iterate over the chunks or args..
        args.insert(4, False)
    else:
        raise RuntimeError(
            f"Unsupported pattern: {pattern}. This is a bug. You shouldn't have ended up on this branch."
        )

    yield from partition(5, args)


def outer_product(args):
    """Scan over a mesh; each motor is on an independent trajectory.

    Parameters
    ----------
    args
        patterned like (``motor1, start1, stop1, num1,```
                        ``motor2, start2, stop2, num2, snake2,``
                        ``motor3, start3, stop3, num3, snake3,`` ...
                        ``motorN, startN, stopN, numN, snakeN``)

        The first motor is the "slowest", the outer loop. For all motors
        except the first motor, there is a "snake" argument: a boolean
        indicating whether to following snake-like, winding trajectory or a
        simple left-to-right trajectory.

    See Also
    --------
    `bluesky.plan_patterns.inner_product`

    Returns
    -------
    cyc : cycler
    """
    shape = []
    extents = []
    snaking = []
    cyclers = []
    for motor, start, stop, num, snake in chunk_outer_product_args(args):
        shape.append(num)
        extents.append([start, stop])
        snaking.append(snake)
        steps = np.linspace(start, stop, num=num, endpoint=True)
        c = cycler(motor, steps)
        cyclers.append(c)

    return snake_cyclers(cyclers, snaking)



"""Extra utilities for state machines, to make them more usable."""

from weakref import WeakKeyDictionary


class ProxyString(str):

    """String that proxies every call to nested machine."""

    def __new__(cls, value, machine):
        """Create new string instance with reference to given machine."""
        string = super(cls, cls).__new__(cls, value)
        string.state_machine = machine
        return string

    def __getattr__(self, name):
        """Proxy call to machine."""
        return getattr(self.state_machine, name)


class PropertyMachine(object):

    """Descriptor to help using machines as properties."""

    def __init__(self, machine_type):
        """Create descriptor."""
        self.memory = WeakKeyDictionary()
        self.machine_type = machine_type

    def __set__(self, instance, value):
        """Set state to machine."""
        self.check_memory(instance)
        self.memory[instance].set_(value)

    def __get__(self, instance, _type=None):
        """Get machine state."""
        if instance is None:
            return self
        self.check_memory(instance)
        machine = self.memory[instance]
        return ProxyString(machine.actual_state.value, machine)

    def check_memory(self, instance):
        try:
            self.memory[instance]
        except KeyError:
            self.memory[instance] = self.machine_type()



from bluesky import Msg
from bluesky.callbacks.olog import logbook_cb_factory

text = []


def f(**kwargs):
    text.append(kwargs["text"])


def test_default_template(RE):
    text.clear()
    RE.subscribe(logbook_cb_factory(f), "start")
    RE([Msg("open_run", plan_args={}), Msg("close_run")])
    assert len(text[0]) > 0


def test_trivial_template(RE):
    text.clear()
    RE.subscribe(logbook_cb_factory(f, desc_template="hello"), "start")
    RE([Msg("open_run", plan_args={}), Msg("close_run")])
    assert text[0] == "hello"

    # smoke test the long_template
    RE.subscribe(logbook_cb_factory(f, long_template="hello"), "start")
    RE([Msg("open_run", plan_args={}), Msg("close_run")])


def test_template_dispatch(RE):
    disp = {"a": "A", "b": "B"}
    text.clear()
    RE.subscribe(logbook_cb_factory(f, desc_dispatch=disp), "start")
    RE([Msg("open_run", plan_name="a", plan_args={}), Msg("close_run")])
    RE([Msg("open_run", plan_name="b", plan_args={}), Msg("close_run")])
    assert text[0] == "A"
    assert text[1] == "B"

    # smoke test the long_dispatch
    RE.subscribe(logbook_cb_factory(f, long_dispatch=disp), "start")
    RE([Msg("open_run", plan_name="a", plan_args={}), Msg("close_run")])
    RE([Msg("open_run", plan_name="b", plan_args={}), Msg("close_run")])



import asyncio
import contextlib
import sys
import tempfile
import threading
from collections import defaultdict


@contextlib.contextmanager
def _print_redirect():
    old_stdout = sys.stdout
    try:
        fout = tempfile.TemporaryFile(mode="w+", encoding="utf-8")
        sys.stdout = fout
        yield fout
    finally:
        sys.stdout = old_stdout


class MsgCollector:
    def __init__(self, msg_hook=None):
        self.msgs = []
        self.msg_hook = msg_hook

    def __call__(self, msg):
        self.msgs.append(msg)
        if self.msg_hook:
            self.msg_hook(msg)


class DocCollector:
    def __init__(self):
        self.start = []
        self.stop = {}
        self.descriptor = defaultdict(list)
        self.event = {}

    def insert(self, name, doc):
        if name == "start":
            self.start.append(doc)
        elif name == "stop":
            self.stop[doc["run_start"]] = doc
        elif name == "descriptor":
            self.descriptor[doc["run_start"]].append(doc)
            self.event[doc["uid"]] = []
        elif name == "bulk_events":
            for k, v in doc.items():
                self.event[k].extend(v)
        else:
            self.event[doc["descriptor"]].append(doc)


def _fabricate_asycio_event(loop):
    th_ev = threading.Event()

    aio_event = None

    def really_make_the_event():
        nonlocal aio_event
        aio_event = asyncio.Event()
        th_ev.set()

    h = loop.call_soon_threadsafe(really_make_the_event)
    if not th_ev.wait(0.1):
        h.cancel()
        raise Exception("failed to make asyncio event")
    return aio_event


def _careful_event_set(ev):
    "Helper to set 'do not lock test suite' backup sets"

    def inner():
        try:
            ev.set()
        except RuntimeError:
            ...

    return inner



import enum

from bluesky._vendor.super_state_machine import extras, machines


class Lock(machines.StateMachine):
    class States(enum.Enum):
        OPEN = "open"
        LOCKED = "locked"

    class Meta:
        allow_empty = False
        initial_state = "open"
        named_transitions = [
            ("lock", "locked"),
            ("open", "open"),
        ]


def test_property_machine():
    class Door:
        lock1 = extras.PropertyMachine(Lock)
        lock2 = extras.PropertyMachine(Lock)

    door = Door()
    assert door.lock1 == "open"
    assert door.lock2 == "open"
    door.lock1.lock()
    assert door.lock1 == "locked"
    assert door.lock2 == "open"
    door.lock2.state_machine.lock()
    assert door.lock1 == "locked"
    assert door.lock2 == "locked"
    door.lock1.open()
    door.lock2.open()
    assert door.lock1 == "open"
    assert door.lock2 == "open"



import json
import os
import uuid
from collections.abc import Iterator
from pathlib import Path
from typing import Optional, Union, cast

import h5py
import jinja2
import numpy as np
import ophyd.sim
import pytest
import tifffile as tf
from event_model.documents.event_descriptor import DataKey
from event_model.documents.stream_datum import StreamDatum
from event_model.documents.stream_resource import StreamResource
from tiled.client import record_history

import bluesky.plan_stubs as bps
import bluesky.plans as bp
from bluesky.callbacks.tiled_writer import TiledWriter
from bluesky.protocols import (
    Collectable,
    HasName,
    Readable,
    Reading,
    StreamAsset,
    WritesStreamAssets,
)

rng = np.random.default_rng(12345)


@pytest.fixture(scope="module")
def catalog(tmp_path_factory):
    tiled_catalog = pytest.importorskip("tiled.catalog")
    tmp_path = tmp_path_factory.mktemp("tiled_catalog")
    yield tiled_catalog.in_memory(
        writable_storage={"filesystem": str(tmp_path), "sql": f"duckdb:///{tmp_path}/test.db"},
        readable_storage=[str(tmp_path.parent)],
    )


@pytest.fixture(scope="module")
def app(catalog):
    tsa = pytest.importorskip("tiled.server.app")
    yield tsa.build_app(catalog)


@pytest.fixture(scope="module")
def context(app):
    tc = pytest.importorskip("tiled.client")
    with tc.Context.from_app(app) as context:
        yield context


@pytest.fixture(scope="module")
def client(context):
    tc = pytest.importorskip("tiled.client")
    yield tc.from_context(context)


@pytest.fixture(scope="module")
def external_assets_folder(tmp_path_factory):
    """External data files used with the saved documents."""
    # Create a temporary directory
    temp_dir = tmp_path_factory.mktemp("example_files")

    # Create an external hdf5 file
    with h5py.File(temp_dir.joinpath("dataset.h5"), "w") as file:
        grp = file.create_group("entry").create_group("data")
        grp.create_dataset("data_1", data=rng.random(size=(3,), dtype="float64"))
        grp.create_dataset("data_2", data=rng.integers(-10, 10, size=(3, 13, 17)), dtype="<i8")

    # Create a second external hdf5 file to be declared in a different stream resource
    with h5py.File(temp_dir.joinpath("dataset_part2.h5"), "w") as file:
        grp = file.create_group("entry").create_group("data")
        grp.create_dataset("data_2", data=rng.integers(-10, 10, size=(5, 13, 17)), dtype="<i8")

    # Create a sequence of tiff files
    (temp_dir / "tiff_files").mkdir(parents=True, exist_ok=True)
    for i in range(3):
        data = rng.integers(0, 255, size=(1, 10, 15), dtype="uint8")
        tf.imwrite(temp_dir.joinpath("tiff_files", f"img_{i:05}.tif"), data)

    yield str(temp_dir.absolute()).replace("\\", "/")


def render_templated_documents(fname: str, data_dir: str):
    dirpath = str(Path(__file__).parent.joinpath("examples"))
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(dirpath), autoescape=False)
    template = env.get_template(fname)
    rendered = template.render(root_path=data_dir, uuid=str(uuid.uuid4())[:-12])

    yield from json.loads(rendered)


class Named(HasName):
    name: str = ""
    root: str = "/root"
    parent = None

    def __init__(self, name: str, root: str) -> None:
        self.name = name
        self.root = root
        self.counter = 0


class StreamDatumReadableCollectable(Named, Readable, Collectable, WritesStreamAssets):
    """Produces no events, but only StreamResources/StreamDatums and can be read or collected"""

    def _get_hdf5_stream(self, data_key: str, index: int) -> tuple[Optional[StreamResource], StreamDatum]:
        file_path = os.path.join(self.root, "dataset.h5")
        uid = f"{data_key}-uid"
        data_desc = self.describe()[data_key]  # Descriptor dictionary for the current data key
        data_shape = cast(tuple[int, ...], tuple(data_desc["shape"]))
        hdf5_dataset = f"/{data_key}/VALUE"

        stream_resource = None
        if self.counter == 0:
            # Backward compatibility test, ignore typing errors
            stream_resource = StreamResource(  # type: ignore[typeddict-unknown-key]
                parameters={"dataset": hdf5_dataset, "chunk_shape": (100, *data_shape[1:]), "_validate": True},
                data_key=data_key,
                root=self.root,
                resource_path="/dataset.h5",
                uri="file://localhost/" + file_path,
                spec="AD_HDF5_SWMR_STREAM",
                mimetype="application/x-hdf5",
                uid=uid,
            )
            # Initialize an empty HDF5 dataset (3D: var 1 dim, fixed 2 and 3 dims)
            with h5py.File(file_path, "a") as f:
                dset = f.require_dataset(
                    hdf5_dataset,
                    data_shape,
                    maxshape=(None, *data_shape[1:]),
                    dtype=np.dtype("float64"),
                    chunks=(100, *data_shape[1:]),
                )

        indx_min, indx_max = self.counter, self.counter + index
        stream_datum = StreamDatum(
            stream_resource=uid,
            descriptor="",
            uid=f"{uid}/{self.counter}",
            indices={"start": indx_min, "stop": indx_max},
            seq_nums={"start": 0, "stop": 0},  # seq_nums will be overwritten by RunBundler
        )

        # Write (append to) the hdf5 dataset
        with h5py.File(file_path, "a") as f:
            dset = f[hdf5_dataset]
            dset.resize([indx_max * data_shape[0], *data_shape[1:]])
            dset[indx_min * data_shape[0] : indx_max * data_shape[0], ...] = np.random.randn(
                (indx_max - indx_min) * data_shape[0], *data_shape[1:]
            )

        return stream_resource, stream_datum

    def _get_tiff_stream(self, data_key: str, index: int) -> tuple[Optional[StreamResource], StreamDatum]:
        file_path = self.root
        for data_key in [f"{self.name}-sd3"]:
            uid = f"{data_key}-uid"
            data_desc = self.describe()[data_key]  # Descriptor dictionary for the current data key
            data_shape = cast(tuple[int, ...], tuple(data_desc["shape"]))
            stream_resource = None
            if self.counter == 0:
                # Backward compatibility test, ignore typing errors
                stream_resource = StreamResource(  # type: ignore[typeddict-unknown-key]
                    parameters={
                        "chunk_shape": (1, *data_shape),
                        "template": "{:05d}.tif",
                        "join_method": "stack",
                        "_validate": True,
                    },
                    data_key=data_key,
                    root=self.root,
                    uri="file://localhost/" + self.root + "/",
                    spec="AD_TIFF",
                    mimetype="multipart/related;type=image/tiff",
                    uid=uid,
                )

            indx_min, indx_max = self.counter, self.counter + index
            stream_datum = StreamDatum(
                stream_resource=uid,
                descriptor="",
                uid=f"{uid}/{self.counter}",
                indices={"start": indx_min, "stop": indx_max},
                seq_nums={"start": 0, "stop": 0},  # seq_nums will be overwritten by RunBundler
            )

            # Write a tiff file
            data = np.random.randint(0, 255, data_shape, dtype="uint8")
            tf.imwrite(os.path.join(file_path, f"{self.counter:05}.tif"), data)

        return stream_resource, stream_datum

    def describe(self) -> dict[str, DataKey]:
        """Describe datasets which will be backed by StreamResources"""
        return {
            # Numerical data with 1 number per event in hdf5 format
            f"{self.name}-sd1": DataKey(
                source="file",
                dtype="number",
                dtype_numpy=np.dtype("float64").str,
                shape=[
                    1,
                ],
                external="STREAM:",
            ),
            # 2-D data with 5 frames per event in hdf5 format
            f"{self.name}-sd2": DataKey(
                source="file",
                dtype="array",
                dtype_numpy=np.dtype("float64").str,
                shape=[5, 10, 15],
                external="STREAM:",
            ),
            # 3-D data with 10 frames per event in tiff format
            f"{self.name}-sd3": DataKey(
                source="file",
                dtype="array",
                dtype_numpy=np.dtype("uint8").str,
                shape=[10, 5, 7, 4],
                external="STREAM:",
            ),
        }

    def describe_collect(self) -> Union[dict[str, DataKey], dict[str, dict[str, DataKey]]]:
        return self.describe()

    def collect_asset_docs(self, index: Optional[int] = None) -> Iterator[StreamAsset]:
        """Produce a StreamResource and StreamDatum for all data keys for 0:index"""
        index = index or 1
        data_keys_methods = {
            f"{self.name}-sd1": self._get_hdf5_stream,
            f"{self.name}-sd2": self._get_hdf5_stream,
            f"{self.name}-sd3": self._get_tiff_stream,
        }

        for data_key, method in data_keys_methods.items():
            stream_resource, stream_datum = method(data_key, index)
            if stream_resource is not None:
                yield "stream_resource", stream_resource
            yield "stream_datum", stream_datum

        self.counter += index

    def get_index(self) -> int:
        """Report how many frames were written"""
        return self.counter

    def read(self) -> dict[str, Reading]:
        """Produce an empty event"""
        return {}


class SynSignalWithRegistry(ophyd.sim.SynSignalWithRegistry):
    """A readable image detector that writes a sequence of files and generates relevant Bluesky documents.

    Subclassed from ophyd.sim to match the updated schema of Resource documents.
    """

    def __init__(self, *args, dtype_numpy="uint8", **kwargs):
        self.dtype_numpy = dtype_numpy
        super().__init__(*args, **kwargs)

    def stage(self):
        super().stage()
        parameters = {
            "chunk_shape": (1,),
            "template": "_{:d}." + self.save_ext,
            "join_method": "stack",
            "_validate": True,
        }
        self._asset_docs_cache[-1][1]["resource_kwargs"].update(parameters)

    def describe(self):
        res = super().describe()
        for key in res:
            res[key]["external"] = "FILESTORE"
            res[key]["dtype_numpy"] = self.dtype_numpy
        return res


def test_stream_datum_readable_counts(RE, client, tmp_path):
    tw = TiledWriter(client)
    det = StreamDatumReadableCollectable(name="det", root=str(tmp_path))
    RE(bp.count([det], 3), tw)
    stream = client.values().last()["primary"]
    keys = sorted(set(stream.base.keys()).difference({"internal"}))

    assert stream[keys[0]].shape == (3,)
    assert stream[keys[1]].shape == (15, 10, 15)
    assert stream[keys[2]].shape == (3, 10, 5, 7, 4)
    assert stream[keys[0]].read() is not None
    assert stream[keys[1]].read() is not None
    assert stream[keys[2]].read() is not None


def test_stream_datum_readable_with_two_detectors(RE, client, tmp_path):
    det1 = StreamDatumReadableCollectable(name="det1", root=str(tmp_path))
    det2 = StreamDatumReadableCollectable(name="det2", root=str(tmp_path))
    tw = TiledWriter(client)
    RE(bp.count([det1, det2], 3), tw)
    stream = client.values().last()["primary"]
    keys = sorted(set(stream.base.keys()).difference({"internal"}))

    assert stream[keys[0]].shape == (3,)
    assert stream[keys[1]].shape == (15, 10, 15)
    assert stream[keys[2]].shape == (3, 10, 5, 7, 4)
    assert stream[keys[3]].shape == (3,)
    assert stream[keys[4]].shape == (15, 10, 15)
    assert stream[keys[5]].shape == (3, 10, 5, 7, 4)
    assert stream[keys[0]].read() is not None
    assert stream[keys[1]].read() is not None
    assert stream[keys[2]].read() is not None
    assert stream[keys[3]].read() is not None
    assert stream[keys[4]].read() is not None
    assert stream[keys[5]].read() is not None


def test_stream_datum_collectable(RE, client, tmp_path):
    det = StreamDatumReadableCollectable(name="det", root=str(tmp_path))
    tw = TiledWriter(client)
    RE(collect_plan(det, name="primary"), tw)
    stream = client.values().last()["primary"]
    keys = sorted(set(stream.base.keys()).difference({"internal"}))

    assert stream[keys[0]].read() is not None
    assert stream[keys[1]].read() is not None
    assert stream[keys[2]].read() is not None


@pytest.mark.parametrize("frames_per_event", [1, 5, 10])
def test_handling_non_stream_resource(RE, client, tmp_path, frames_per_event):
    det = SynSignalWithRegistry(
        func=lambda: np.random.randint(0, 255, (frames_per_event, 10, 15), dtype="uint8"),
        dtype_numpy=np.dtype("uint8").str,
        name="img",
        labels={"detectors"},
        save_func=tf.imwrite,
        save_path=str(tmp_path),
        save_spec="AD_TIFF",
        save_ext="tif",
    )
    tw = TiledWriter(client)
    RE(bp.count([det], 3), tw)
    extr = client.values().last()["primary"].base["img"]
    intr = client.values().last()["primary"].base["internal"]
    assert extr.shape == (3, frames_per_event, 10, 15)
    assert extr.read() is not None
    assert set(intr.columns) == {"seq_num", "time"}
    assert len(intr.read()) == 3
    assert (intr["seq_num"].read() == [1, 2, 3]).all()


def collect_plan(*objs, name="primary"):
    yield from bps.open_run()
    yield from bps.declare_stream(*objs, collect=True, name=name)
    yield from bps.collect(*objs, return_payload=False, name=name)
    yield from bps.close_run()


@pytest.mark.parametrize("fname", ["internal_events", "external_assets", "external_assets_legacy"])
@pytest.mark.parametrize("batch_size", [0, 1, 1000, None])
def test_with_correct_sample_runs(client, batch_size, external_assets_folder, fname):
    if batch_size is None:
        tw = TiledWriter(client)
    else:
        tw = TiledWriter(client, batch_size=batch_size)
    for item in render_templated_documents(fname + ".json", external_assets_folder):
        if item["name"] == "start":
            uid = item["doc"]["uid"]
        tw(**item)

    run = client[uid]

    for stream in run.values():
        assert stream.read() is not None


@pytest.mark.parametrize("error_type", ["shape", "chunks", "dtype"])
@pytest.mark.parametrize("validate", [True, False])
def test_validate_external_data(client, external_assets_folder, error_type, validate):
    tw = TiledWriter(client)

    documents = render_templated_documents("external_assets_single_key.json", external_assets_folder)
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the document to introduce an error
        if (error_type == "shape") and (name == "descriptor"):
            doc["data_keys"]["det-key2"]["shape"] = [1, 2, 3]  # should be [1, 13, 17]
        elif (error_type == "chunks") and name in {"resource", "stream_resource"}:
            doc["parameters"]["chunk_shape"] = [1, 2, 3]  # should be [100, 13, 17]
        elif (error_type == "dtype") and (name == "descriptor"):
            doc["data_keys"]["det-key2"]["dtype_numpy"] = np.dtype("int32").str  # should be "int64"

        # Add flag to trigger validation
        if name in {"resource", "stream_resource"} and validate:
            doc["parameters"]["_validate"] = True

        # Check that the warning is issued when data changes during the validation
        if name == "stop" and validate:
            with pytest.warns(UserWarning):
                tw(name, doc)
        else:
            tw(name, doc)

    # Try reading the imported data
    run = client[uid]
    if not validate and not error_type == "chunks":
        with pytest.raises(ValueError):
            assert run["primary"].read() is not None
    else:
        assert run["primary"].read() is not None
        assert run["primary"]["det-key2"].read().shape == (8, 13, 17)


@pytest.mark.parametrize("squeeze", [True, False])
def test_slice_and_squeeze(client, external_assets_folder, squeeze):
    tw = TiledWriter(client)

    documents = render_templated_documents("external_assets_single_key.json", external_assets_folder)
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the documents to add slice and squeeze parameters
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["shape"] = [1, 17] if squeeze else [1, 5, 17]
        elif name in {"resource", "stream_resource"}:
            doc["parameters"]["slice"] = ":,5,:" if squeeze else ":,:5,:"
            doc["parameters"]["squeeze"] = squeeze
            doc["parameters"]["chunk_shape"] = [1]

        tw(name, doc)

    # Try reading the imported data
    assert client[uid]["primary"].read() is not None


def test_legacy_multiplier_parameter(client, external_assets_folder):
    tw = TiledWriter(client)

    documents = render_templated_documents("external_assets_single_key.json", external_assets_folder)
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the documents to add slice and squeeze parameters
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["shape"] = [13, 17]
        elif name in {"resource", "stream_resource"}:
            doc["parameters"]["multiplier"] = 1

        tw(name, doc)

    # Try reading the imported data
    assert client[uid]["primary"].read() is not None


def test_streams_with_no_events(client, external_assets_folder):
    tw = TiledWriter(client)

    for item in render_templated_documents("external_assets_single_key.json", external_assets_folder):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Skip the resource and datum documents
        if name in {"resource", "stream_resource", "datum", "stream_datum", "event"}:
            continue

        tw(name, doc)

    # Try reading the data -- should return an empty dataset
    assert client[uid]["primary"].read() is not None
    assert client[uid]["primary"].read().data_vars == {}
    assert client[uid]["primary"].metadata is not None


@pytest.mark.parametrize("include_data_sources", [True, False])
@pytest.mark.parametrize("fname", ["internal_events", "external_assets", "external_assets_legacy"])
def test_zero_gets(client, external_assets_folder, fname, include_data_sources):
    client = client.new_variation(include_data_sources=include_data_sources)
    assert client._include_data_sources == include_data_sources
    tw = TiledWriter(client)
    assert bool(tw.client._include_data_sources)

    with record_history() as history:
        for item in render_templated_documents(fname + ".json", external_assets_folder):
            tw(**item)

    # Count the number of GET requests
    num_gets = sum(1 for req in history.requests if req.method == "GET")
    assert num_gets == 0


def test_bad_document_order(client, external_assets_folder):
    """Test that the TiledWriter can handle documents in a different order than expected

    Emit datum documents in the end, before the Stop document, but after corresponding Event documents.
    """
    tw = TiledWriter(client)

    document_cache = []
    for item in render_templated_documents("external_assets_legacy.json", external_assets_folder):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        if name == "datum":
            document_cache.append({"name": name, "doc": doc})
            continue

        if name == "stop":
            for cached_item in document_cache:
                tw(**cached_item)

        tw(**item)

    run = client[uid]

    for stream in run.values():
        assert stream.read() is not None
        assert "time" in stream.keys()
        assert "seq_num" in stream.keys()
        assert len(stream.keys()) > 2  # There's at least one data key in addition to time and seq_num


def test_json_backup(client, tmpdir, monkeypatch):
    def patched_event(name, doc):
        raise RuntimeError("This is a test error to check the backup functionality")

    monkeypatch.setattr("bluesky.callbacks.tiled_writer._RunWriter.event", patched_event)

    tw = TiledWriter(client, backup_directory=str(tmpdir))

    for item in render_templated_documents("internal_events.json", ""):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]
        print(name)

        tw(**item)

    run = client[uid]

    assert "primary" in run  # The Descriptor was processed and the primary stream was created
    assert run["primary"].read() is not None  # The stream can be read
    assert len(run["primary"].read()) == 0  # No events were processed due to the error
    assert "stop" in run.metadata  # The TiledWriter did not crash

    # Check that the backup file was created
    filepath = tmpdir / f"{uid[:8]}.jsonl"
    assert filepath.exists()
    with open(filepath) as f:
        lines = [json.loads(line) for line in f if line.strip()]
    assert len(lines) == 7
    assert lines[0]["name"] == "start"
    assert lines[1]["name"] == "descriptor"
    assert lines[2]["name"].startswith("event")
    assert lines[6]["name"] == "stop"



import uuid
from collections.abc import Generator
from functools import partial
from math import isclose
from time import time
from typing import Any
from unittest.mock import MagicMock

import pytest

import bluesky.plan_stubs as bps
import bluesky.plans as bp
from bluesky import Msg
from bluesky.plans import grid_scan, scan
from bluesky.simulators import (
    END,
    RunEngineSimulator,
    assert_message_and_return_remaining,
    check_limits,
    plot_raster_path,
    print_summary,
    print_summary_wrapper,
    summarize_plan,
)


def test_print_summary(hw):
    det = hw.det
    motor = hw.motor
    print_summary(scan([det], motor, -1, 1, 10))  # old name
    summarize_plan(scan([det], motor, -1, 1, 10))  # new name
    list(print_summary_wrapper(scan([det], motor, -1, 1, 10)))


def test_old_module_name(hw):
    det = hw.det
    motor = hw.motor
    motor1 = hw.motor1
    motor2 = hw.motor2
    from bluesky.plan_tools import (
        plot_raster_path,
        print_summary,
        print_summary_wrapper,
    )

    with pytest.warns(UserWarning):
        print_summary(scan([det], motor, -1, 1, 10))
    with pytest.warns(UserWarning):
        list(print_summary_wrapper(scan([det], motor, -1, 1, 10)))
    with pytest.warns(UserWarning):
        plan = grid_scan([det], motor1, -5, 5, 10, motor2, -7, 7, 15, True)
        plot_raster_path(plan, "motor1", "motor2", probe_size=0.3)


def test_check_limits(RE, hw):
    det = hw.det
    motor = hw.motor
    # The motor object does not currently implement limits.
    # Use an assert to help us out if this changes in the future.
    assert not hasattr(motor, "limits")

    # # check_limits should warn if it can't find check_value
    # TODO: Is there _any_ object to test?
    # with pytest.warns(UserWarning):
    #     check_limits(scan([det], motor, -1, 1, 3))

    # monkey-patch some limits
    motor.limits = (-2, 2)
    # check_limits should do nothing here
    check_limits(scan([det], motor, -1, 1, 3))

    # check_limits should error if limits are exceeded only if object raises
    # this object does not raise
    check_limits(scan([det], motor, -3, 3, 3))

    # check_limits should raise if limits are equal only if object raises
    # this object does not raise
    motor.limits = (2, 2)
    check_limits(scan([det], motor, -1, 1, 3))


def test_check_limits_needs_RE():
    with pytest.raises(RuntimeError) as ctx:
        check_limits([])
    assert str(ctx.value) == "Bluesky event loop not running"


def test_plot_raster_path(hw):
    det = hw.det
    motor1 = hw.motor1
    motor2 = hw.motor2
    plan = grid_scan([det], motor1, -5, 5, 10, motor2, -7, 7, 15, True)
    plot_raster_path(plan, "motor1", "motor2", probe_size=0.3)


def test_simulator_simulates_simple_plan():
    def simple_plan() -> Generator[Msg, Any, Any]:
        yield from bps.null()

    sim = RunEngineSimulator()
    messages = sim.simulate_plan(simple_plan())
    assert len(messages) == 1
    assert messages[0] == Msg("null")


def test_simulator_add_handler(hw):
    callback = MagicMock()

    def do_sleep():
        yield from bps.sleep("60")

    sim = RunEngineSimulator()
    sim.add_handler("sleep", lambda msg: callback(msg))
    msgs = sim.simulate_plan(do_sleep())
    callback.assert_called_once()
    msg = callback.mock_calls[0].args[0]
    assert msg.command == "sleep"
    assert len(msgs) == 1
    assert msgs[0].command == "sleep"


@pytest.mark.parametrize(
    "insert_order, expected",
    [([0, 0, 0], 2), ([END, END, END], 0), ([0, 0, 1], 1), ([0, -1, 2], 1)],
)
def test_simulator_add_handler_append_with_index(hw, insert_order, expected):
    sim = RunEngineSimulator()

    def generate_value(id, msg):
        return {"values": {"value": id}}

    for index, id in zip(insert_order, range(0, len(insert_order))):
        sim.add_handler("read", partial(generate_value, id), None, index)

    def read_det():
        value = yield from bps.rd("det_a")
        return value

    sim.simulate_plan(read_det())
    result = sim.return_value
    assert result == expected


def test_simulator_add_read_handler_for(hw):
    def trigger_and_return_position():
        yield from bps.trigger(hw.ab_det)
        pos = yield from bps.rd(hw.ab_det.a)
        return pos

    sim = RunEngineSimulator()
    sim.add_read_handler_for(hw.ab_det.a, 5)
    msgs = sim.simulate_plan(trigger_and_return_position())
    assert sim.return_value == 5
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "trigger" and msg.obj.name == "det"
    )
    assert_message_and_return_remaining(msgs, lambda msg: msg.command == "read" and msg.obj.name == "det_a")


def test_simulator_add_read_handler_for_with_dict(hw):
    def trigger_and_return_position():
        yield from bps.trigger(hw.ab_det)
        pos = yield from bps.rd(hw.ab_det.a)
        return pos

    sim = RunEngineSimulator()
    sim.add_read_handler_for(hw.ab_det.a, {"value": 5})
    msgs = sim.simulate_plan(trigger_and_return_position())
    assert sim.return_value == 5
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "trigger" and msg.obj.name == "det"
    )
    assert_message_and_return_remaining(msgs, lambda msg: msg.command == "read" and msg.obj.name == "det_a")


def test_simulator_add_read_handler_for_multiple(hw):
    def read_all_values():
        data = yield from bps.read(hw.ab_det)
        return data

    sim = RunEngineSimulator()
    sim.add_read_handler_for_multiple(hw.ab_det, a=11, b={"value": 12, "timestamp": 1717071719})
    sim.simulate_plan(read_all_values())
    assert sim.return_value["a"]["value"] == 11
    assert sim.return_value["b"]["value"] == 12
    assert isclose(sim.return_value["a"]["timestamp"], time(), abs_tol=1)
    assert sim.return_value["b"]["timestamp"] == 1717071719


def test_simulator_add_wait_handler(hw):
    def trigger_and_return_position():
        pos = yield from bps.rd(hw.ab_det.a)
        assert pos == 0
        yield from bps.trigger(hw.ab_det, wait=True)
        pos = yield from bps.rd(hw.ab_det.a)
        return pos

    sim = RunEngineSimulator()
    sim.add_read_handler_for(hw.ab_det.a, 0)
    sim.add_wait_handler(lambda _: sim.add_read_handler_for(hw.ab_det.a, 5))
    msgs = sim.simulate_plan(trigger_and_return_position())
    assert sim.return_value == 5
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "read" and msg.obj.name == "det_a")
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "trigger" and msg.obj.name == "det"
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")
    assert_message_and_return_remaining(msgs, lambda msg: msg.command == "read" and msg.obj.name == "det_a")


def test_fire_callback(hw):
    callback = MagicMock()
    run_start_uid = uuid.uuid4()
    descriptor_uid = uuid.uuid4()
    event_uid = uuid.uuid4()

    def take_a_reading():
        yield from bps.subscribe("all", callback)
        yield from bp.count([hw.det], num=1)

    start_doc = {"plan_name": "count", "uid": run_start_uid}
    descriptor_doc = {"run_start": run_start_uid, "uid": descriptor_uid}
    event_doc = {"descriptor": descriptor_uid, "uid": event_uid}

    sim = RunEngineSimulator()
    sim.add_handler_for_callback_subscribes()

    sim.add_callback_handler_for(
        "open_run",
        "start",
        start_doc,
        lambda msg: msg.kwargs["plan_name"] == "count",
    )
    sim.add_callback_handler_for_multiple("save", [[("descriptor", descriptor_doc), ("event", event_doc)]])

    sim.simulate_plan(take_a_reading())
    calls = callback.mock_calls
    assert len(calls) == 3
    assert calls[0].args == ("start", start_doc)
    assert calls[1].args == ("descriptor", descriptor_doc)
    assert calls[2].args == ("event", event_doc)


def test_assert_message_and_return_remaining(hw):
    def take_three_readings():
        yield from bp.count([hw.det], num=3)

    sim = RunEngineSimulator()
    msgs = sim.simulate_plan(take_three_readings())
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "stage" and msg.obj.name == "det")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and msg.kwargs["plan_name"] == "count"
        and msg.kwargs["num_points"] == 3,
    )
    for _ in range(0, 3):
        msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "checkpoint")
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "trigger" and msg.obj.name == "det"
        )
        trigger_group = msgs[0].kwargs["group"]
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "wait" and msg.kwargs["group"] == trigger_group,  # noqa: B023
        )
        msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "create")
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "read" and msg.obj.name == "det"
        )
        msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "save")
    assert msgs[1].command == "close_run"
